print("import: 'pymatgen'")
import pymatgen

print("import: 'pymatgen.alchemy'")
import pymatgen.alchemy

print("import: 'pymatgen.analysis'")
import pymatgen.analysis

print("import: 'pymatgen.analysis.chemenv'")
import pymatgen.analysis.chemenv

print("import: 'pymatgen.analysis.chemenv.coordination_environments'")
import pymatgen.analysis.chemenv.coordination_environments

print("import: 'pymatgen.analysis.chemenv.coordination_environments.coordination_geometries_files'")
import pymatgen.analysis.chemenv.coordination_environments.coordination_geometries_files

print("import: 'pymatgen.analysis.chemenv.coordination_environments.strategy_files'")
import pymatgen.analysis.chemenv.coordination_environments.strategy_files

print("import: 'pymatgen.analysis.chemenv.utils'")
import pymatgen.analysis.chemenv.utils

print("import: 'pymatgen.analysis.cost'")
import pymatgen.analysis.cost

print("import: 'pymatgen.analysis.defects'")
import pymatgen.analysis.defects

print("import: 'pymatgen.analysis.diffraction'")
import pymatgen.analysis.diffraction

print("import: 'pymatgen.analysis.elasticity'")
import pymatgen.analysis.elasticity

print("import: 'pymatgen.analysis.hhi'")
import pymatgen.analysis.hhi

print("import: 'pymatgen.analysis.pourbaix_diagram'")
import pymatgen.analysis.pourbaix_diagram

print("import: 'pymatgen.apps'")
import pymatgen.apps

print("import: 'pymatgen.apps.battery'")
import pymatgen.apps.battery

print("import: 'pymatgen.apps.borg'")
import pymatgen.apps.borg

print("import: 'pymatgen.cli'")
import pymatgen.cli

print("import: 'pymatgen.command_line'")
import pymatgen.command_line

print("import: 'pymatgen.core'")
import pymatgen.core

print("import: 'pymatgen.electronic_structure'")
import pymatgen.electronic_structure

print("import: 'pymatgen.entries'")
import pymatgen.entries

print("import: 'pymatgen.io'")
import pymatgen.io

print("import: 'pymatgen.io.abinit'")
import pymatgen.io.abinit

print("import: 'pymatgen.io.feff'")
import pymatgen.io.feff

print("import: 'pymatgen.io.lammps'")
import pymatgen.io.lammps

print("import: 'pymatgen.io.vasp'")
import pymatgen.io.vasp

print("import: 'pymatgen.ext.matproj'")
import pymatgen.ext.matproj

print("import: 'pymatgen.optimization'")
import pymatgen.optimization

print("import: 'pymatgen.analysis.phase_diagram'")
import pymatgen.analysis.phase_diagram

print("import: 'pymatgen.analysis.structure_prediction'")
import pymatgen.analysis.structure_prediction

print("import: 'pymatgen.symmetry'")
import pymatgen.symmetry

print("import: 'pymatgen.transformations'")
import pymatgen.transformations

print("import: 'pymatgen.util'")
import pymatgen.util

print("import: 'pymatgen.vis'")
import pymatgen.vis

