# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

"""
This package contains classes to parse input files from the exciting
code package.
"""

from .inputs import ExcitingInput
