# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

"""
The transformations package defines various transformations that can be applied
on structures, i.e., converting one structure to another.
"""
